open Encodings
open Schema_pcre

type node =
  ('a Pxp_document.node Pxp_document.extension as 'a) Pxp_document.node

module Node = struct
  type t = node
  let compare = Pxp_document.compare
end

let start_with s pr =
  let s = Utf8.get_str s in
  (String.length s >= String.length pr) &&
    (String.sub s 0 (String.length pr) = pr)

let has_xsd_prefix s = start_with s "xsd:"

let xsd_namespace = Utf8.mk "http://www.w3.org/2001/XMLSchema"
let xsi_namespace = Utf8.mk "http://www.w3.org/2001/XMLSchema-instance"
let xsd_prefix = Utf8.mk "xsd"
let xsi_prefix = Utf8.mk "xsi"

let schema_ns_prefixes =
  [ xsd_prefix, xsd_namespace; xsi_prefix, xsi_namespace ]

let spec = Pxp_tree_parser.default_namespace_spec
let new_xsd_config () =
  let ns_manager = new Pxp_dtd.namespace_manager in
  List.iter
    (fun (p, ns) -> ns_manager#add_namespace (Utf8.get_str p) (Utf8.get_str ns))
    schema_ns_prefixes;
  { Pxp_types.default_namespace_config with      
      Pxp_types.encoding = `Enc_utf8;
      Pxp_types.enable_namespace_processing = Some ns_manager
  }

let node_of src =
  (Pxp_tree_parser.parse_wfdocument_entity (new_xsd_config ()) src spec) # root

let node_of_uri uri =
  try
    let source = match Url.process uri with 
      | Url.Filename s -> Pxp_types.from_file s 
      | Url.Url s -> Pxp_types.from_string s
    in
    node_of source
  with exn -> raise (Location.Generic (Pxp_types.string_of_exn exn))

let _may_attr name n =
  try
    match n#attribute name with
      | Pxp_types.Value v -> Some (Utf8.mk v)
      | _ -> None
  with Not_found -> None
    
let _is_attr name n v =
  try
    match n#attribute name with
      | Pxp_types.Value v' -> v = v'
      | _ -> false
  with Not_found -> false
    
let _attr name n =
  match n#attribute name with
    | Pxp_types.Value v -> Utf8.mk v
    | _ -> failwith ("Attribute " ^ name ^ " is missing")
	
let _may_elem e (n: node) =
  try Some (Pxp_document.find_element e n) with Not_found -> None
    
let _elems e n = Pxp_document.find_all_elements e n

let _filter_elems p n =
  Pxp_document.find_all (fun n -> match n#node_type with
			   | Pxp_document.T_element s -> List.mem s p
			   | _ -> false) n
  
let _line n = match n#position with (_,l,_) -> l
  
let _iter_nodes n f = n#iter_nodes f
  
let _iter_elems n f = n#iter_nodes
  (fun n ->
     match n#node_type with
       | Pxp_document.T_element s -> f n s
       | _ -> ()
  )

let _fold_elems n x f = 
  let x = ref x in
  n#iter_nodes
    (fun n ->
       match n#node_type with
	 | Pxp_document.T_element s -> x := f !x n s
	 | _ -> ()
    );
  !x
  
let _tag n =
  match n#node_type with
    | Pxp_document.T_element s -> s
    | _ -> assert false
	
let _has_tag n f =
  match n#node_type with
    | Pxp_document.T_element s -> f s
    | _ -> false
	
let _namespaces n =
  List.map 
    (fun n ->
       (match n#node_type with 
	    Pxp_document.T_namespace p -> p | _ -> assert false),
       n#data
    )
    n#namespaces_as_nodes
    
let _find p n = Pxp_document.find p n

let _resolve_qname n qname =
  let (prefix,local) = Ns.split_qname qname in
  let ns = 
    try Ns.mk (Utf8.mk (n # namespace_scope # uri_of_display_prefix prefix))
    with Not_found -> Ns.empty
  in
  (ns,local)

let _may_qname_attr name n =
  match _may_attr name n with
    | Some qname -> Some (_resolve_qname n qname)
    | None -> None

let _qname_attr name n =
  match _may_attr name n with
    | Some qname -> _resolve_qname n qname
    | None -> assert false


let xsd = Ns.mk xsd_namespace
let xsi = Ns.mk xsi_namespace
